@@ -28,6 +28,8 @@ from __future__ import absolute_import
 
 import six
 import inspect
+import binascii
+import random
 
 import txaio
 txaio.use_twisted()  # noqa
@@ -45,9 +47,11 @@ from autobahn.twisted.rawsocket import WampRawSocketClientFactory
 from autobahn.websocket.compress import PerMessageDeflateOffer, \
     PerMessageDeflateResponse, PerMessageDeflateResponseAccept
 
-from autobahn.wamp import protocol
+from autobahn.wamp import protocol, auth
+from autobahn.wamp.interfaces import IAuthenticator
 from autobahn.wamp.types import ComponentConfig
 
+
 __all__ = [
     'ApplicationSession',
     'ApplicationSessionFactory',
@@ -114,7 +118,12 @@ class ApplicationRunner(object):
                  serializers=None,
                  ssl=None,
                  proxy=None,
-                 headers=None):
+                 headers=None,
+                 max_retries=None,
+                 initial_retry_delay=None,
+                 max_retry_delay=None,
+                 retry_delay_growth=None,
+                 retry_delay_jitter=None):
         """
 
         :param url: The WebSocket URL of the WAMP router to connect to (e.g. `ws://somehost.com:8090/somepath`)
@@ -144,6 +153,21 @@ class ApplicationRunner(object):
 
         :param headers: Additional headers to send (only applies to WAMP-over-WebSocket).
         :type headers: dict
+
+        :param max_retries: Maximum number of reconnection attempts. Unlimited if set to -1.
+        :type max_retries: int
+
+        :param initial_retry_delay: Initial delay for reconnection attempt in seconds (Default: 1.0s).
+        :type initial_retry_delay: float
+
+        :param max_retry_delay: Maximum delay for reconnection attempts in seconds (Default: 60s).
+        :type max_retry_delay: float
+
+        :param retry_delay_growth: The growth factor applied to the retry delay between reconnection attempts (Default 1.5).
+        :type retry_delay_growth: float
+
+        :param retry_delay_jitter: A 0-argument callable that introduces nose into the delay. (Default random.random)
+        :type retry_delay_jitter: float
         """
         assert(type(url) == six.text_type)
         assert(realm is None or type(realm) == six.text_type)
@@ -157,6 +181,11 @@ class ApplicationRunner(object):
         self.ssl = ssl
         self.proxy = proxy
         self.headers = headers
+        self.max_retries = max_retries
+        self.initial_retry_delay = initial_retry_delay
+        self.max_retry_delay = max_retry_delay
+        self.retry_delay_growth = retry_delay_growth
+        self.retry_delay_jitter = retry_delay_jitter
 
         # this if for auto-reconnection when Twisted ClientService is avail
         self._client_service = None
@@ -324,19 +353,26 @@ class ApplicationRunner(object):
             # this code path is automatically reconnecting ..
             self.log.debug('using t.a.i.ClientService')
 
-            default_retry = backoffPolicy()
+            if self.max_retries or self.initial_retry_delay or self.max_retry_delay or self.retry_delay_growth or self.retry_delay_jitter:
+                kwargs = {}
+                for key, val in [('initialDelay', self.initial_retry_delay),
+                                 ('maxDelay', self.max_retry_delay),
+                                 ('factor', self.retry_delay_growth),
+                                 ('jitter', lambda: random.random() * self.retry_delay_jitter)]:
+                    if val:
+                        kwargs[key] = val
 
-            if False:
                 # retry policy that will only try to reconnect if we connected
                 # successfully at least once before (so it fails on host unreachable etc ..)
                 def retry(failed_attempts):
-                    if self._connect_successes > 0:
-                        return default_retry(failed_attempts)
+                    if self._connect_successes > 0 and (self.max_retries == -1 or failed_attempts < self.max_retries):
+                        return backoffPolicy(**kwargs)(failed_attempts)
                     else:
+                        print('hit stop')
                         self.stop()
                         return 100000000000000
             else:
-                retry = default_retry
+                retry = backoffPolicy()
 
             self._client_service = ClientService(client, transport_factory, retryPolicy=retry)
             self._client_service.startService()
@@ -723,27 +759,12 @@ if service:
 
 
 # new API
-class Session(ApplicationSession):
-    # shim that lets us present pep8 API for user-classes to override,
-    # but also backwards-compatible for existing code using
-    # ApplicationSession "directly".
+class Session(protocol._SessionShim):
+    # XXX these methods are redundant, but put here for possibly
+    # better clarity; maybe a bad idea.
 
-    # XXX note to self: if we release this as "the" API, then we can
-    # change all internal Autobahn calls to .on_join() etc, and make
-    # ApplicationSession a subclass of Session -- and it can then be
-    # separate deprecated and removed, ultimately, if desired.
-
-    def onJoin(self, details):
-        return self.on_join(details)
-
-    # XXX def onChallenge(self, )
-    # XXX def onOpen ??
-
-    def onLeave(self, details):
-        return self.on_leave(details)
-
-    def onDisconnect(self):
-        return self.on_disconnect()
+    def on_welcome(self, welcome_msg):
+        pass
 
     def on_join(self, details):
         pass
@@ -751,5 +772,92 @@ class Session(ApplicationSession):
     def on_leave(self, details):
         self.disconnect()
 
+    def on_connect(self):
+        self.join(self.config.realm)
+
     def on_disconnect(self):
         pass
+
+
+# experimental authentication API
+class AuthCryptoSign(object):
+
+    def __init__(self, **kw):
+        # should put in checkconfig or similar
+        for key in kw.keys():
+            if key not in [u'authextra', u'authid', u'authrole', u'privkey']:
+                raise ValueError(
+                    "Unexpected key '{}' for {}".format(key, self.__class__.__name__)
+                )
+        for key in [u'privkey', u'authid']:
+            if key not in kw:
+                raise ValueError(
+                    "Must provide '{}' for cryptosign".format(key)
+                )
+        for key in kw.get('authextra', dict()):
+            if key not in [u'pubkey']:
+                raise ValueError(
+                    "Unexpected key '{}' in 'authextra'".format(key)
+                )
+
+        from autobahn.wamp.cryptosign import SigningKey
+        self._privkey = SigningKey.from_key_bytes(
+            binascii.a2b_hex(kw[u'privkey'])
+        )
+
+        if u'pubkey' in kw.get(u'authextra', dict()):
+            pubkey = kw[u'authextra'][u'pubkey']
+            if pubkey != self._privkey.public_key():
+                raise ValueError(
+                    "Public key doesn't correspond to private key"
+                )
+        else:
+            kw[u'authextra'] = kw.get(u'authextra', dict())
+            kw[u'authextra'][u'pubkey'] = self._privkey.public_key()
+        self._args = kw
+
+    def on_challenge(self, session, challenge):
+        return self._privkey.sign_challenge(session, challenge)
+
+
+IAuthenticator.register(AuthCryptoSign)
+
+
+class AuthWampCra(object):
+
+    def __init__(self, **kw):
+        # should put in checkconfig or similar
+        for key in kw.keys():
+            if key not in [u'authextra', u'authid', u'authrole', u'secret']:
+                raise ValueError(
+                    "Unexpected key '{}' for {}".format(key, self.__class__.__name__)
+                )
+        for key in [u'secret', u'authid']:
+            if key not in kw:
+                raise ValueError(
+                    "Must provide '{}' for wampcra".format(key)
+                )
+
+        self._args = kw
+        self._secret = kw.pop(u'secret')
+        if not isinstance(self._secret, six.text_type):
+            self._secret = self._secret.decode('utf8')
+
+    def on_challenge(self, session, challenge):
+        key = self._secret.encode('utf8')
+        if u'salt' in challenge.extra:
+            key = auth.derive_key(
+                key,
+                challenge.extra['salt'],
+                challenge.extra['iterations'],
+                challenge.extra['keylen']
+            )
+
+        signature = auth.compute_wcs(
+            key,
+            challenge.extra['challenge'].encode('utf8')
+        )
+        return signature.decode('ascii')
+
+
+IAuthenticator.register(AuthWampCra)
